/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenixpro/configs/Configs.hpp"
#include "ctre/phoenixpro/controls/ControlRequest.hpp"
#include "ctre/phoenixpro/networking/interfaces/Control_Interface.h"
#include <sstream>
#include <units/angle.h>
#include <units/current.h>
#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenixpro {
namespace controls {

/**
 * Request PID to target position with duty-cycle feed forward.
 *
 * This control mode will set the motor's position setpoint to the position specified by the user. In
 * addition, it will apply an additional torque current as an arbitrary feed-forward value.
 */
class PositionTorqueCurrentFOC : public ControlRequest
{
    bool ApplyConfigsOnRequest;

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;
        auto& ref = requestReference.GetNameValues();
        ss << Position.to<double>(); ref["Position"] = ss.str(); ss.str(std::string());
        ss << FeedForward.to<double>(); ref["FeedForward"] = ss.str(); ss.str(std::string());
        ss << Slot; ref["Slot"] = ss.str(); ss.str(std::string());
        ss << OverrideCoastDurNeutral; ref["OverrideCoastDurNeutral"] = ss.str(); ss.str(std::string());

        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<PositionTorqueCurrentFOC *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<PositionTorqueCurrentFOC>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenixpro_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenixpro_RequestControlPositionTorqueCurrentFOC(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests, Position.to<double>(), FeedForward.to<double>(), Slot, OverrideCoastDurNeutral);
    }

public:
    /**
     * Position to drive toward in rotations.
     */
    units::angle::turn_t Position;
    /**
     * Feedforward to apply in torque current in Amperes.
     */
    units::current::ampere_t FeedForward;
    /**
     * Select which gains are applied by selecting the slot.  Use the configuration
     * api to set the gain values for the selected slot before enabling this
     * feature. Slot must be within [0,2].
     */
    int Slot;
    /**
     * Set to true to coast the rotor when output is zero (or within deadband).  Set
     * to false to use the NeutralMode configuration setting (default). This flag
     * exists to provide the fundamental behavior of this control when output is
     * zero, which is to provide 0A (zero torque).
     */
    bool OverrideCoastDurNeutral;

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{100_Hz}; // Default to 100_Hz

    /**
     *\brief  Request PID to target position with duty-cycle feed forward.
     *
     *\details  This control mode will set the motor's position setpoint to the
     *          position specified by the user. In addition, it will apply an
     *          additional torque current as an arbitrary feed-forward value.
     *
     * \param Position     Position to drive toward in rotations.
     * \param FeedForward     Feedforward to apply in torque current in Amperes.
     * \param Slot     Select which gains are applied by selecting the slot. 
     *                 Use the configuration api to set the gain values for the
     *                 selected slot before enabling this feature. Slot must be
     *                 within [0,2].
     * \param OverrideCoastDurNeutral     Set to true to coast the rotor when
     *                                    output is zero (or within deadband). 
     *                                    Set to false to use the NeutralMode
     *                                    configuration setting (default). This
     *                                    flag exists to provide the fundamental
     *                                    behavior of this control when output
     *                                    is zero, which is to provide 0A (zero
     *                                    torque).
     */
    PositionTorqueCurrentFOC(units::angle::turn_t Position, units::current::ampere_t FeedForward, int Slot, bool OverrideCoastDurNeutral) : ControlRequest{"PositionTorqueCurrentFOC"}, ApplyConfigsOnRequest{false}
    {
        this->Position = Position;
        this->FeedForward = FeedForward;
        this->Slot = Slot;
        this->OverrideCoastDurNeutral = OverrideCoastDurNeutral;
    }

    /**
     *\brief  Request PID to target position with duty-cycle feed forward.
     *
     *\details  This control mode will set the motor's position setpoint to the
     *          position specified by the user. In addition, it will apply an
     *          additional torque current as an arbitrary feed-forward value.
     *
     * \param Position     Position to drive toward in rotations.
     */
    PositionTorqueCurrentFOC(units::angle::turn_t Position) : PositionTorqueCurrentFOC{Position, 0.0_A, 0, false}
    {}
    
    /**
     * \brief Modifies this Control Request's Position parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newPosition Parameter to modify
     * \returns Itself
     */
    PositionTorqueCurrentFOC& WithPosition(units::angle::turn_t newPosition)
    {
        Position = newPosition;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's FeedForward parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newFeedForward Parameter to modify
     * \returns Itself
     */
    PositionTorqueCurrentFOC& WithFeedForward(units::current::ampere_t newFeedForward)
    {
        FeedForward = newFeedForward;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's Slot parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newSlot Parameter to modify
     * \returns Itself
     */
    PositionTorqueCurrentFOC& WithSlot(int newSlot)
    {
        Slot = newSlot;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's OverrideCoastDurNeutral parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newOverrideCoastDurNeutral Parameter to modify
     * \returns Itself
     */
    PositionTorqueCurrentFOC& WithOverrideCoastDurNeutral(bool newOverrideCoastDurNeutral)
    {
        OverrideCoastDurNeutral = newOverrideCoastDurNeutral;
        return *this;
    }
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    PositionTorqueCurrentFOC &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: PositionTorqueCurrentFOC" << std::endl;
        ss << "Position: " << Position.to<double>() << std::endl;
        ss << "FeedForward: " << FeedForward.to<double>() << std::endl;
        ss << "Slot: " << Slot << std::endl;
        ss << "OverrideCoastDurNeutral: " << OverrideCoastDurNeutral << std::endl;
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}

